from abc import ABC, abstractmethod
from dataclasses import dataclass
import json
import logging
from typing import Any, Dict, List, Literal, Optional, Tuple, Union, cast
from anthropic import Anthropic
from anthropic.types import MessageParam
import requests
from openai import AzureOpenAI, OpenAI, OpenAIError
import jsonpath_ng

# Exceptions
from .exceptions import Uncontactable, status_code_to_exception, openai_exception_to_exception, EmptyResponse, NotImplemented
from .headers import parse_headers
from .responses import extract_reply, extract_replies
from .types import type_model_presets


@dataclass
class PromptResponse:
    prompt:str
    response:str

class Context:
    def __init__(self) -> None:
        self.turns:list[PromptResponse] = []

    def add(self, prompt_response:PromptResponse) -> None:
        self.turns.append(prompt_response)

class ContextManager:
    def __init__(self) -> None:
        self.contexts:Dict[str,Context] = {}

    def get_context_or_none(self, context_id:Optional[str] = None) -> Optional[Context]:
        if context_id is None:
            return None
        if self.contexts.get(context_id, None) is None:
            self.contexts[context_id] = Context()
        return self.contexts[context_id]

class ModelWrapper(ABC):
    @abstractmethod
    def __call__(self, content:str, with_context:Optional[Context] = None) -> str:
        pass

class TestStaticResponder(ModelWrapper):
    """
    This is only for testing
    """
    def __init__(self, system_prompt: str):
        self.context_manager = ContextManager()
        self._system_prompt = system_prompt
        pass
    
    def __call__(self, content:str, with_context:Optional[Context] = None) -> str:
        request = f"[start]sys: {self._system_prompt};"
        if with_context is not None:
            request = f"{request}"
            for prompt_response in with_context.turns:
                request = f"{request} user: {prompt_response.prompt}; assistant: {prompt_response.response};"
            request = f"{request}"
        else:
            request = f"{request}"
        request = f"{request} next: {content}[end]"

        response = f"TEST. prompted with: {request=}"

        if with_context is not None:
            with_context.add(PromptResponse(
                prompt=content,
                response=response
            ))
        return response

class APIModelWrapper(ModelWrapper):
    def __init__(
        self,
        api_url: str,
        request_template: Optional[str] = None,
        selector: Optional[str] = None,
        system_prompt: Optional[str] = None,
        tokenizer: Optional[str] = None,
        headers: Optional[Dict[str, str]] = None
    ) -> None:
        self.context_manager = ContextManager()
        self.api_url = api_url
        if tokenizer:
            logging.debug("Note that with tokenizer enabled, the request_template format is different.")
            self.request_template =  request_template or '{"prompt": "{tokenized_chat_template}"}'
            if '{tokenized_chat_template}' not in self.request_template:
                raise ValueError("`--request-template` must contain '{tokenized_chat_template}' when using a tokenizer.")
            from transformers import AutoTokenizer
            self.tokenizer = AutoTokenizer.from_pretrained(tokenizer)
        else:
            self.tokenizer = None
            default_template = '{"prompt": "{system_prompt}{prompt}"}' if system_prompt is None else '{"prompt": "{system_prompt} {prompt}"}'
            self.request_template = request_template or default_template
            if '{prompt}' not in self.request_template or '{system_prompt}' not in self.request_template:
                raise ValueError("`--request-template` must contain '{prompt}' and '{system_prompt}'.")
        self.selector = selector
        self.system_prompt = system_prompt or ""
        self.headers = headers or {}


    def prompt_to_request_payload(self, prompt: str) -> Dict[str, Any]:
        if self.tokenizer:
            tokenized_chat_template = cast(str, self.tokenizer.apply_chat_template([
                {"role": "system", "content": self.system_prompt},
                {"role": "user", "content": prompt}
            ], tokenize=False))

            tokenized_chat_template = json.dumps(tokenized_chat_template, ensure_ascii=False)
            payload = self.request_template.replace("{tokenized_chat_template}", tokenized_chat_template)
        else:
            # Dump to escape quote marks that are inside the prompt/system_prompt
            prompt = json.dumps(prompt, ensure_ascii=False)
            system_prompt = json.dumps(self.system_prompt, ensure_ascii=False)
            # The dumps added extra quote marks to prompt and system prompt so trim them before the replace
            payload = self.request_template.replace("{prompt}", prompt[1:-1])
            payload = payload.replace("{system_prompt}", system_prompt[1:-1])

        logging.debug(f"{payload=}")
        # should handle non-json payload (or single string)
        payload = json.loads(payload)
        assert isinstance(payload, dict), f"Expected the request template to form a json dict, got {type(payload)} instead."
        return cast(Dict[str, Any], payload)


    def __call__(self, content:str, with_context:Optional[Context] = None) -> str:
        if with_context is not None:
            logging.debug("APIModelWrapper is temporarily incompatible with chat completions history. Attacks that require chat completions history fail.")
            raise NotImplemented("The crescendo attack is currently incompatible with custom model wrappers.")

        request_payload = self.prompt_to_request_payload(content)

        # Make the API call
        try:
            response = requests.post(self.api_url, headers=self.headers, json=request_payload)
            response.raise_for_status()
        except requests.exceptions.ConnectionError as cerr:
            raise Uncontactable(str(cerr))
        except requests.exceptions.HTTPError as httperr:
            status_code: int = httperr.response.status_code
            raise status_code_to_exception(status_code)
        except Exception as e:
            # everything else
            raise e

        return extract_replies(response, self.selector)


class AzureAIStudioWrapper(APIModelWrapper):
    def __init__(self, api_key: str, url: str, request_template: Optional[str], system_prompt: str) -> None:
        az_request_template = request_template or """{
    "messages": [
        {"role": "system", "content": "{system_prompt}"},
        {"role": "user", "content": "{prompt}"}
    ],
    "temperature": 0.2,
    "max_tokens": 1024,
    "stop": [],
    "top_p": 0.9
}"""
        super().__init__(
            url,
            request_template=az_request_template,
            selector='["choices"][0]["message"]["content"]',
            headers={"Authorization": f'Bearer {api_key}', "Content-Type": "application/json"},
            system_prompt=system_prompt
        )
    
    def __call__(self, content:str, with_context:Optional[Context] = None) -> str:
        if with_context is not None:
            logging.debug("APIModelWrapper is temporarily incompatible with chat completions history. Attacks that require chat completions history fail.")
            raise NotImplementedError("The crescendo attack is currently incompatible with custom model wrappers.")

        request_payload = self.prompt_to_request_payload(content)

        # Make the API call
        response = requests.post(self.api_url, headers=self.headers, json=request_payload)
        if response.status_code == 400:
            try:
                # Detect Azure Content Filter error message
                err_res_json = response.json()
                if err_message := err_res_json.get("error", {}).get("message", None):
                    return cast(str, err_message)
            except Exception:
                raise status_code_to_exception(400)
        elif response.status_code != 200:
            # Handle other types of API error
            message = f"API call failed with {response.status_code=} {response.json()=}."
            raise status_code_to_exception(response.status_code)

        res_json: Dict[str, Any] = cast(Dict[str, Any], response.json())

        # Detect Cohere content_filter error
        if res_json["choices"][0]["finish_reason"] == "content_filter":
            return "Sorry, but a content filter was triggered. Please try again with a different prompt."
        
        # Now we are sure that the response was successful and not a content filter error
        if self.selector:
            jsonpath_expr = jsonpath_ng.parse(self.selector)
            match = jsonpath_expr.find(res_json)
            if match:
                return str(match[0].value)
            else:
                raise Exception(f"Selector {self.selector} did not match any elements in the response. {res_json=}")

        # disabled until we can support templating chat completions
        # if with_context is not None:
        #     with_context.add(PromptResponse(
        #         prompt=content,
        #         response=response
        #     ))
    
        return str(response)


class HuggingFaceWrapper(APIModelWrapper):
    def __init__(self, api_key: str, api_url: str, request_template: str, system_prompt: Optional[str] = None) -> None:
        super().__init__(
            api_url,
            request_template=request_template,
            selector='[0]["generated_text"]',
            headers={'Authorization': f'Bearer {api_key}'},
            system_prompt=system_prompt
        )


# https://learn.microsoft.com/en-us/azure/ai-services/openai/reference#chat-completions to see available values for az_api_version
# ["2023-05-15", "2023-06-01-preview", "2023-10-01-preview", "2024-02-15-preview", "2024-03-01-preview", "2024-04-01-preview", "2024-02-01"] on May 15th 2024
class AzureOpenAIWrapper(ModelWrapper):
    def __init__(self, api_key: str, model_name: str, az_api_version: str, url: str, system_prompt: Optional[str] = None) -> None:
        self.api_key = api_key
        self.model_name = model_name
        self.client = AzureOpenAI(api_key=api_key, api_version=az_api_version, azure_endpoint=url)
        self.system_prompt = system_prompt

    def __call__(self, content:str, with_context:Optional[Context] = None) -> str:
        return openai_call(wrapper=self, content=content, with_context=with_context)


class OpenAIWrapper(ModelWrapper):
    def __init__(self, api_key: str, model_name: Optional[str], system_prompt: Optional[str] = None, api_url: Optional[str] = None) -> None:
        self.api_key = api_key
        self.client = OpenAI(api_key=api_key) if api_url is None else OpenAI(api_key=api_key, base_url=api_url)
        self.model_name = model_name or "gpt-3.5-turbo"
        self.system_prompt = system_prompt

    def __call__(self, content:str, with_context:Optional[Context] = None) -> str:
        return openai_call(wrapper=self, content=content, with_context=with_context)


def openai_call(wrapper: Union[AzureOpenAIWrapper, OpenAIWrapper], content:str, with_context:Optional[Context] = None) -> str:
    if wrapper.system_prompt:
        messages = [{"role": "system", "content": wrapper.system_prompt}]
    else:
        messages = []

    if with_context:
        for prompt_response in with_context.turns:
            messages.append({"role":"user","content": prompt_response.prompt})
            messages.append({"role":"assistant","content": prompt_response.response})
    
    messages.append({"role":"user", "content":content})

    logging.debug(messages)

    try:
        chat = wrapper.client.chat.completions.create(model=wrapper.model_name, messages=messages)  # type: ignore # TODO: fix type error
        response = chat.choices[0].message.content
        if not response:
            raise EmptyResponse("Model returned an empty response.")
    except OpenAIError as e:
        raise openai_exception_to_exception(e)

    if with_context is not None:
        with_context.add(PromptResponse(
            prompt=content,
            response=response,
        ))
    return response


class AnthropicWrapper(ModelWrapper):
    def __init__(self, api_key: str, model_name: Optional[str], system_prompt: Optional[str] = None) -> None:
        self.api_key = api_key
        self.client = Anthropic(api_key=api_key)
        self.model_name = model_name or "claude-3-opus-20240229"
        self.system_prompt = system_prompt

    def __call__(self, content:str, with_context:Optional[Context] = None) -> str:
        messages: List[MessageParam] = []

        messages = []
        if with_context:
            for prompt_response in with_context.turns:
                messages.append({"role":"user","content": prompt_response.prompt})
                messages.append({"role":"assistant","content": prompt_response.response})

        messages.append({"role":"user", "content":content})

        if self.system_prompt is not None:
            message = self.client.messages.create(
                system=self.system_prompt,
                max_tokens=1024, 
                messages=messages, 
                model=self.model_name
            )
        else:
            message = self.client.messages.create(
                max_tokens=1024, 
                messages=messages, 
                model=self.model_name
            )
        response = message.content[0].text

        if with_context is not None:
            with_context.add(PromptResponse(
                prompt=content,
                response=response,
            ))
        return response
    

    

def check_expected_args(args: Dict[str, Any], expected_args: List[str]) -> None:
    missing_args: List[str] = []
    for arg in expected_args:
        if not args.get(arg):
            missing_args.append(f"`--{arg.replace('_', '-')}`")
    if missing_args:
        raise ValueError(f"Missing required arguments: {', '.join(missing_args)}")


def get_model_wrapper(
    headers: Dict[str,str],
    preset: Optional[type_model_presets] = None,
    api_key: Optional[str] = None,
    url: Optional[str] = None,
    model_name: Optional[str] = None,
    az_api_version: Optional[str] = None,
    system_prompt: Optional[str] = None,
    selector: Optional[str] = None,
    request_template: Optional[str] = None,
    tokenizer: Optional[str] = None
) -> ModelWrapper:

    # Create model based on preset
    if preset == 'huggingface-openai':
        check_expected_args(locals(), ['api_key', 'url'])
        api_key, url, request_template = cast(Tuple[str, str, str], (api_key, url, request_template))
        if url[-4:] != "/v1/":
            url += "/v1/"
        return OpenAIWrapper(api_key=api_key, api_url=url, system_prompt=system_prompt, model_name="tgi")
    elif preset == "huggingface":
        check_expected_args(locals(), ['api_key', 'url', 'request_template'])
        api_key, url, request_template = cast(Tuple[str, str, str], (api_key, url, request_template))
        return HuggingFaceWrapper(api_key=api_key, api_url=url, system_prompt=system_prompt, request_template=request_template)
    elif preset == 'azure-aistudio':
        check_expected_args(locals(), ['api_key', 'url', 'system_prompt'])
        api_key, url, system_prompt = cast(Tuple[str, str, str], (api_key, url, system_prompt))
        return AzureAIStudioWrapper(api_key=api_key, url=url, request_template=request_template, system_prompt=system_prompt)
    elif preset == 'openai':
        check_expected_args(locals(), ['api_key'])
        api_key = cast(str, api_key)
        return OpenAIWrapper(api_key=api_key, model_name=model_name, system_prompt=system_prompt)
    elif preset == 'azure-openai':
        check_expected_args(locals(), ['api_key', 'model_name', 'az_api_version', 'url'])
        api_key, model_name, az_api_version, url = cast(Tuple[str, str, str, str], (api_key, model_name, az_api_version, url))
        return AzureOpenAIWrapper(api_key=api_key, model_name=model_name, az_api_version=az_api_version, url=url, system_prompt=system_prompt)
    elif preset == 'anthropic':
        if not api_key:
            raise ValueError("`--api-key` argument is required when using the 'anthropic' preset.")
        return AnthropicWrapper(api_key=api_key, model_name=model_name, system_prompt=system_prompt)
    elif preset == 'tester':
        if not system_prompt:
            raise ValueError("`--system-prompt` argument is required")
        return TestStaticResponder(system_prompt=system_prompt)
    else:
        if not url:
            raise ValueError("`--url` argument is required when not using a preset configuration.")
        # Convert headers string to dictionary
        if headers:
            return APIModelWrapper(api_url=url, selector=selector, request_template=request_template, system_prompt=system_prompt, tokenizer=tokenizer, headers=headers)
        else:
            return APIModelWrapper(api_url=url, selector=selector, request_template=request_template, system_prompt=system_prompt, tokenizer=tokenizer)


def parse_args_into_model(args: Dict[str, Any]) -> ModelWrapper:
    return get_model_wrapper(
        preset=args["preset"],
        headers=parse_headers(headers_comma_separated=args["headers"], headers_list=args["header"]),
        api_key=args["api_key"],
        url=args["url"],
        selector=args["selector"],
        request_template=args["request_template"],
        system_prompt=args["system_prompt"],
        model_name=args["model_name"],
        az_api_version=args["az_api_version"],
        tokenizer=args["tokenizer"],
    )


# Exceptions
from typing import Tuple
from .exceptions import Uncontactable, HTTPBaseError

# Models
from .wrappers.llm import LLMModelWrapper
from .wrappers.image import ImageModelWrapper

# UI
from rich.console import Console

# Logging
import logging


# Data
from .utils import base64_test_image_as_bytes


def preflight_llm(
    model_wrapper: LLMModelWrapper,
    console: Console,
    json_out: bool,
) -> bool:
    """
    Makes requests to the LLM to validate basic connectivity before submitting
    test.

    Returns True on success, False on failure
    """
    try:
        if not json_out:
            console.print("[white]Contacting model...")
        for i in range(5):
            _ = model_wrapper.__call__("Hello llm, are you there?")
        return True
    except Uncontactable as cerr:
        logging.debug(cerr)
        model_api = (
            model_wrapper.api_url if hasattr(model_wrapper, "api_url") else "<unknown>"
        )
        if not json_out:
            console.print(
                f"[red]Could not connect to the model! [white](URL: {model_api}, are you sure it's correct?)"
            )
    except HTTPBaseError as httpbe:
        logging.debug(httpbe)
        message: str = (
            f"[red]Model pre-flight check returned {httpbe.status_code} ({httpbe.status_message})"
        )
        if not json_out:
            console.print(message)
    except Exception as e:
        # something we've not really accounted for caught
        logging.error(e)
        raise e

    return False


def preflight_image(
    model_wrapper: ImageModelWrapper,
    console: Console,
    json_out: bool,
) -> bool:
    
    try:
        data = base64_test_image_as_bytes()   
        for i in range(5):
            _ = model_wrapper.__call__(data)
        return True
    except Exception as e:
        if not json_out:
            console.print(f"[red]Could not contact the model!")
        logging.error(e)

    return False


import time

# Auth
from .auth import require_auth

# Types
from typing import Any

# UI
from rich.progress import Progress, SpinnerColumn, TaskID, TextColumn
from rich.table import Table
from rich.live import Live
from rich.console import Console

# Type aliases
from .types import (
    type_submit_func,
    type_polling_func,
    type_output_func,
    type_ui_task_map,
    type_ui_exception_map,
)


def output_placeholder(_: str, __: bool) -> None:
    return None


@require_auth
def cli_run(
    submit_func: type_submit_func,
    polling_func: type_polling_func,
    # access_token cannot be the first arg it seems
    access_token: str,
    json_out: bool,
    submitting_text: str = "Submitting...",
    output_func: type_output_func = output_placeholder,
) -> Any:

    submit_progress = Progress(
        "{task.description}",
        SpinnerColumn(finished_text="[green3] Submitted!"),
        auto_refresh=False,
        disable=json_out,
    )

    progress_table = Table.grid(expand=True)
    overall_task_progress = Progress(disable=json_out)

    ui_task_progress = Progress(
        "{task.description}",
        SpinnerColumn(finished_text="done"),
        TextColumn("{task.fields[status]}"),
        disable=json_out,
    )
    ui_task_map: type_ui_task_map = {}

    ui_exceptions_progress = Progress("{task.description}", disable=json_out)
    ui_exception_map: type_ui_exception_map = {}

    with submit_progress:
        task_id = submit_progress.add_task(submitting_text, start=True)
        initial_result = submit_func(
            access_token, ui_exception_map, ui_exceptions_progress
        )
        submit_progress.update(task_id, completed=100)

    if not json_out:
        progress_table.add_row(overall_task_progress)
        progress_table.add_row(ui_task_progress)
        progress_table.add_row("")
        progress_table.add_row(ui_exceptions_progress)

    polled_result = polling_func(
        access_token, initial_result, ui_task_map, ui_task_progress
    )
    tasks_to_run = len(ui_task_progress.tasks)
    # don't show progress bar if there is only one task
    if tasks_to_run > 1:
        overall_task_id = overall_task_progress.add_task("Progress", total=tasks_to_run)

    with Live(progress_table, refresh_per_second=10):
        while polled_result is None:
            polled_result = polling_func(
                access_token, initial_result, ui_task_map, ui_task_progress
            )
            if tasks_to_run > 1:
                completed_tasks = sum(task.completed for task in ui_task_progress.tasks)
                overall_task_progress.update(overall_task_id, completed=completed_tasks)
            time.sleep(1)

    output_table = output_func(polled_result, json_out)
    if output_table:
        Console().print(output_table)

    return polled_result



# Typing
import json
from typing import Optional
from .run_poll_display import type_ui_task_map

# Orchestrator
from .orchestrator import OrchestratorTestResponse, get_test_by_id

# Constants
from .constants import DASHBOARD_URL

# UI
from rich.progress import Progress
from rich.table import Table


def poll_and_display_test(
    access_token: str,
    ui_task_map: type_ui_task_map,
    ui_task_progress: Progress,
    initial_test: OrchestratorTestResponse,
) -> Optional[OrchestratorTestResponse]:
    test = get_test_by_id(access_token=access_token, test_id=initial_test.id)

    if len(ui_task_map.keys()) == 0:
        for attack in test.attacks:
            ui_task_map[attack.id] = ui_task_progress.add_task(
                f"Attack {attack.attack}", total=1, status="[chartreuse1]queued"
            )

    for attack in test.attacks:
        task_id = ui_task_map[attack.id]
        if attack.state == 2:
            ui_task_progress.update(task_id, completed=1, status="[chartreuse3]success")
        elif attack.state == -1:
            ui_task_progress.update(task_id, completed=1, status="[red3]failed")
        elif attack.state == 1:
            ui_task_progress.update(task_id, status="[orange3]running")

    if test.hasFinished is False:
        return None
    return test


def output_test_table(
    json_out: bool,
    test: OrchestratorTestResponse,
    risk_threshold: int,
) -> Optional[Table]:
    if json_out:
        print(test.model_dump_json())
        return None
    else:
        table = Table(title=f"Results - {DASHBOARD_URL}/r/test/{test.id}", width=80)
        table.add_column("Pass", style="cyan")
        table.add_column("Name", style="magenta")
        table.add_column("Risk", justify="right", style="green")

        for attack in test.attacks:
            if attack.state != 2:
                name = f"Error running '{attack.attack}'"
                risk_str = "n/a"
                emoji = "❗️"
            else:
                name = attack.attack
                risk_str = str(attack.risk)
                emoji = "❌‍" if attack.risk > risk_threshold else "✅️"

            table.add_row(emoji, name, risk_str)

        return table



from ..types import (
    type_ui_exception_map,
    ExceptionCountTuple,
)

from typing import Optional, Callable, Literal, Optional, Dict, Type, Any

from rich.progress import Progress

import time

import logging

# Networking
from azure.messaging.webpubsubclient.models import OnGroupDataMessageArgs
from azure.messaging.webpubsubclient import WebPubSubClient

from ..wrappers.llm import LLMModelWrapper, ContextManager

# Exceptions
from ..exceptions import (
    MGException,
    ServiceUnavailable,
    InternalServerError,
    RateLimitOrInsufficientCredits,
    FailedDependency,
    UnprocessableEntity,
    Timeout,
    NotFound,
    BadRequest,
    Forbidden,
    Unauthorized,
    Uncontactable,
    HTTPBaseError,
    EmptyResponse,
    NotImplemented,
)

ErrorCode = Literal[
    "CouldNotContact",
    "ContentPolicy",
    "CLIError",
    "NotImplemented",
    "NoResponse",
    "RateLimited",
    "NetworkIssue",
    "MaxContextLength",
]

exceptions_to_cli_status_codes: Dict[Type[Exception], ErrorCode] = {
    Uncontactable: "CouldNotContact",
    BadRequest: "NoResponse",  # not sure about this, we don't handle 400 atm
    Unauthorized: "NoResponse",
    Forbidden: "NoResponse",
    NotFound: "NoResponse",
    Timeout: "NoResponse",
    UnprocessableEntity: "NoResponse",  # this is currently being handled as a rate limit issue for some reason
    FailedDependency: "NoResponse",
    RateLimitOrInsufficientCredits: "RateLimited",
    InternalServerError: "NoResponse",
    ServiceUnavailable: "NoResponse",
    NotImplemented: "NotImplemented",
    EmptyResponse: "NoResponse",
}


def handle_exception_callback(
    exception: Exception,
    handle_visual_exception_callback: Optional[Callable[[str], None]],
) -> ErrorCode:
    # TODO - come take a look at this
    error_code: ErrorCode = exceptions_to_cli_status_codes.get(type(exception), "CLIError")  # type: ignore
    callback_text = str(exception)

    if handle_visual_exception_callback:
        handle_visual_exception_callback(callback_text)

    logging.debug(exception)
    return error_code


def llm_message_handler(
    ui_exception_map: type_ui_exception_map,
    ui_exception_progress: Progress,
    submitted_test_id_update: Callable[[str], None],
    model_wrapper: LLMModelWrapper,
    ws_client: WebPubSubClient,
) -> Callable[[OnGroupDataMessageArgs], None]:
    def _handle_visual_exception_callback(text: str) -> None:
        if not ui_exception_map.get(text, None):
            ui_exception_map[text] = ExceptionCountTuple(
                ui_exception_progress.add_task(""), 0
            )

        ui_exception_map[text].count += 1
        ui_exception_progress.update(
            ui_exception_map[text].task_id,
            description=f"[dark_orange3][!!!] {text} (x{ui_exception_map[text].count})",
        )

    def temp_handler(e: Exception) -> ErrorCode:
        return handle_exception_callback(e, _handle_visual_exception_callback)

    context_manager = ContextManager()

    def recv_message_handler(msg: OnGroupDataMessageArgs) -> None:
        if msg.data["messageType"] == "Request":
            logging.debug(f"received request {msg.data=}")
            context_id = msg.data["payload"].get("context_id", None)
            context = context_manager.get_context_or_none(context_id)
            content = msg.data["payload"]["prompt"]

            response: str = ""
            error_code: Optional[ErrorCode] = None

            try:
                # would pose being explicit with __call__ so we can ctrl+f easier, not a very clear shorthand
                response = model_wrapper.__call__(
                    content=content,
                    with_context=context,
                )
            except MGException as mge:
                error_code = temp_handler(mge)
                if error_code == "CLIError":
                    raise mge
            except Exception as e:
                raise e
            finally:
                # we always try to send a response
                replyData = {
                    "correlationId": msg.data["correlationId"],
                    "messageType": "Response",
                    "status": "ok",  # Deprecated but kept for compatibility
                    "payload": {
                        "response": response,
                        "error": error_code,
                    },
                }
                logging.debug(f"sending response {replyData=}")

                ws_client.send_to_group("orchestrator", replyData, data_type="json")  # type: ignore
        elif (
            msg.data["messageType"] == "StartedTest"
        ):  # should be something like "Submitted", upstream change required.
            submitted_test_id_update(msg.data["payload"]["testId"])
        else:
            pass

    return recv_message_handler



from typing import List, Optional

from ..types import (
    type_ui_exception_map,
    type_ui_task_map,
)

from rich.progress import Progress

from ..api_service import api_get
from ..orchestrator import get_tests, OrchestratorTestResponse

from ..utils import print_to_stderr_as_json

from rich.table import Table

from datetime import datetime

from ..constants import DASHBOARD_URL


def list_test_submit(
    access_token: str,
    ui_exception_map: type_ui_exception_map,
    ui_exception_progress: Progress,
) -> List[OrchestratorTestResponse]:
    tests_res = get_tests(access_token, request_function=api_get)

    return tests_res


def list_test_polling(
    access_token: str,
    tests_response: List[OrchestratorTestResponse],
    ui_task_map: type_ui_task_map,
    ui_task_progress: Progress,
) -> Optional[List[OrchestratorTestResponse]]:
    return tests_response


def list_test_output(
    test_list: List[OrchestratorTestResponse], json_out: bool
) -> Optional[Table]:
    if json_out:
        print_to_stderr_as_json([test.model_dump() for test in test_list])
        return None
    else:
        table = Table(title=f"Tests run by user")
        table.add_column("Model Name", style="cyan")
        table.add_column("Assessment Date", style="magenta")
        table.add_column("Risk Score", justify="right", style="green")
        table.add_column("URL", justify="right", style="green")

        for test in test_list:
            table.add_row(
                test.mindgardModelName,
                datetime.fromisoformat(test.createdAt).strftime("%H:%M, %Y-%m-%d"),
                str(test.risk),
                f"{DASHBOARD_URL}/r/test/{test.id}",
            )
        return table



"""
Status: Pre-Alpha -- implementation incomplete and interfaces are likely to change

Provides headless execution of mindgard tests.
"""

from dataclasses import dataclass
import logging
from threading import Condition
import time
from typing import Any, Dict, Optional, Protocol, Tuple
from azure.messaging.webpubsubclient import WebPubSubClient, WebPubSubClientCredential
from azure.messaging.webpubsubclient.models import OnGroupDataMessageArgs, CallbackType, WebPubSubDataType
import requests
from mindgard.wrappers.image import ImageModelWrapper

from mindgard.version import VERSION
from mindgard.wrappers.llm import LLMModelWrapper

class RequestHandler(Protocol):
    """
    Protocol for converting a Request paylod to a Response payload
    """
    def __call__(self, payload: Any) -> Any: ...

@dataclass
class ModelConfig:
    def to_orchestrator_init_params(self) -> Dict[str, Any]:
        return {}


@dataclass
class ImageModelConfig(ModelConfig):
    wrapper: ImageModelWrapper
    dataset: str
    labels: list[str]
    model_type: str = "image"

    def to_orchestrator_init_params(self) -> Dict[str, Any]:
        return {
            "modelType": self.model_type,
            "dataset": self.dataset,
            "labels": self.labels,
        }

@dataclass
class LLMModelConfig(ModelConfig):
    wrapper: LLMModelWrapper
    system_prompt: str
    model_type: str = "llm"

    def to_orchestrator_init_params(self) -> Dict[str, Any]:
        return {
            "modelType": self.model_type,
            "system_prompt": self.system_prompt,
        }

@dataclass
class TestConfig:
    api_base: str
    api_access_token: str
    target: str
    attack_source: str
    parallelism: int
    model: ModelConfig
    attack_pack: str = "sandbox"
    additional_headers: Optional[Dict[str, str]] = None
    def to_orchestrator_init_params(self) -> Dict[str, Any]:
        """
        Get parameters for the init test request to orchestrator
        """
        return {**{
            "target": self.target,
            "attackPack": self.attack_pack,
            "parallelism": self.parallelism,
            "attackSource": self.attack_source
        }, **self.model.to_orchestrator_init_params()}

    def handler(self) -> Any:
        return self.model.wrapper.to_handler() # type: ignore # TODO not sure on type


class TestImplementationProvider():

    def init_test(self, config:TestConfig) -> Tuple[str, str]:
        """
        Init a test in with API and return the url and group_id
        """
        url = f"{config.api_base}/tests/cli_init"

        response = requests.post(
            url=url, 
            headers={
                "Authorization": f"Bearer {config.api_access_token}",
                "User-Agent": f"mindgard-cli/{VERSION}",
                "X-User-Agent": f"mindgard-cli/{VERSION}",
                **(config.additional_headers or {})
            },
            json=config.to_orchestrator_init_params()
        )
        response.raise_for_status()

        url = response.json().get("url", None)
        group_id = response.json().get("groupId", None)
        if not url or not group_id:
            raise ValueError("Invalid response from orchestrator, missing websocket credentials.")

        return url, group_id
    def create_client(self, connection_url:str) -> WebPubSubClient:
        credentials = WebPubSubClientCredential(client_access_url_provider=connection_url)
        return WebPubSubClient(credential=credentials)

    def connect_websocket(self, client:WebPubSubClient) -> None:
        client.open()

    def register_handler(self, handler:RequestHandler, client:WebPubSubClient, group_id:str) -> None:
        def callback(msg:OnGroupDataMessageArgs) -> None:
            if msg.data["messageType"] != "Request":
                return
            
            payload = handler(payload=msg.data["payload"])

            client.send_to_group(
                "orchestrator",
                {
                    "correlationId": msg.data["correlationId"],
                    "messageType": "Response",
                    "status": "ok",
                    "payload": payload
                },
                data_type=WebPubSubDataType.JSON
            )
            
        client.subscribe(CallbackType.GROUP_MESSAGE, callback)

    def start_test(self, client:WebPubSubClient, group_id:str) -> str:
        logging.error(f"Starting test {group_id}")
        started_condition = Condition()
        test_ids: list[str] = []
        def handler(msg:OnGroupDataMessageArgs) -> None:
            logging.error(f"received message {msg.data}")
            if msg.data["messageType"] == "StartedTest":
                with started_condition:
                    test_ids.append(msg.data["payload"]["testId"])
                    started_condition.notify_all()

        client.subscribe(CallbackType.GROUP_MESSAGE, handler)
        payload = {
            "correlationId": "",
            "messageType": "StartTest",
            "payload": {"groupId": group_id},
        }
        client.send_to_group(group_name="orchestrator", content=payload, data_type=WebPubSubDataType.JSON)
        
        logging.error("waiting for test to start")
        with started_condition:
            started_condition.wait_for(lambda: len(test_ids) > 0)
            return test_ids[0]

    def poll_test(self, config:TestConfig, test_id:str, period_seconds:int = 5) -> None:
        finished = False
        while not finished:
            response = requests.get(
                url=f"{config.api_base}/assessments/{test_id}",
                headers={
                    "Authorization": f"Bearer {config.api_access_token}",
                    **(config.additional_headers or {})
                }
            )
            try:
                if response.status_code == 200:
                    test = response.json()
                    finished = test["hasFinished"]
                    logging.info(f"Test {test_id} has finished! {test['hasFinished']} {test['isCompleted']}")
            except requests.JSONDecodeError as jde:
                logging.error(f"Error decoding response: {jde}")
                pass
            except KeyError as ke:
                logging.error(f"KeyError response: {ke}")
                pass
                
            time.sleep(period_seconds)

    def close(self, client:Optional[WebPubSubClient]) -> None:
        if client:
            client.close()
            
class Test:
    def __init__(self, config:TestConfig, provider:TestImplementationProvider = TestImplementationProvider()):
        self._config = config
        self._provider = provider

    def run(self) -> None:
        p = self._provider
        wps_client = None
        try:
            wps_url, group_id = p.init_test(self._config)
            logging.info("Creating webpubsub client")
            wps_client = p.create_client(wps_url)
            logging.info("Connecting to webpubsub client")
            p.connect_websocket(wps_client)
            handler = self._config.handler()
            p.register_handler(handler, wps_client, group_id)
            logging.info("Submitting test")
            test_id = p.start_test(wps_client, group_id)
            logging.info("Polling test...")
            p.poll_test(self._config, test_id)
            logging.info("...Test complete!")
        finally:
            logging.info("Closing webpubsub client")
            p.close(wps_client)


from .config import load_auth_config
 
VERSION: str = "0.53.0"
REPOSITORY_URL: str = f"https://pypi.org/pypi/mindgard/json"

ALGORITHMS = ['RS256']

auth_config = load_auth_config()
AUTH0_DOMAIN = auth_config.AUTH0_DOMAIN
AUTH0_CLIENT_ID = auth_config.AUTH0_CLIENT_ID
AUTH0_AUDIENCE = auth_config.AUTH0_AUDIENCE
API_BASE = auth_config.API_BASE
DASHBOARD_URL = auth_config.DASHBOARD_URL

API_RETRY_ATTEMPTS = 10
API_RETRY_WAIT_BETWEEN_ATTEMPTS_SECONDS = 3



import jsonpath_ng
import json

# Filters to only LLM messages with content.
MINIMUM_MESSAGE_SIZE = 10

def extract_reply(json_response,selector=None, strict=True):
    if selector:
        jsonpath_expr = jsonpath_ng.parse(selector)
        match = jsonpath_expr.find(json_response)
        if match:
            return str(match[0].value)
        elif strict:
            raise Exception(f"Selector {selector} did not match any elements in the response. {json_response=}")
        else:
            return ""
    else:
        return json.dumps(json_response)

def extract_replies(response, selector=None):
    # Example of spec: https://platform.openai.com/docs/api-reference/streaming
    if (response.headers.get('Content-Type','').lower() == 'text/event-stream'):
        reply = []
        for line in response.iter_lines():
            line_value = line[len('data: '):].decode("utf-8")
            if len(line_value) < MINIMUM_MESSAGE_SIZE:
                continue
            line_json = json.loads(line_value)
            extracted = extract_reply(line_json,selector=selector, strict=False).strip()
            if (len(extracted) > 0):
                reply.append(extracted)

        return " ".join(reply)
    else:
        # Simple RPC
        return extract_reply(response.json(), selector=selector)


from typing import Any, Dict, Literal, Union

from mindgard.headers import parse_headers
from mindgard.wrappers.image import ImageModelWrapper, get_image_model_wrapper
from mindgard.wrappers.llm import LLMModelWrapper, get_llm_model_wrapper


# TODO: make an alias for model_type
def parse_args_into_model(
    model_type: Literal["image", "llm"], args: Dict[str, Any]
) -> Union[LLMModelWrapper, ImageModelWrapper]:
    if model_type == "llm":
        return get_llm_model_wrapper(
            preset=args["preset"],
            headers=parse_headers(headers_comma_separated=args["headers"], headers_list=args["header"]),
            api_key=args["api_key"],
            url=args["url"],
            selector=args["selector"],
            request_template=args["request_template"],
            system_prompt=args["system_prompt"],
            model_name=args["model_name"],
            az_api_version=args["az_api_version"],
            tokenizer=args["tokenizer"],
        )
    elif model_type == "image":
        return get_image_model_wrapper(
            preset=args["preset"], api_key=args["api_key"], url=args["url"], labels=args["labels"]
        )



import argparse
import sys
import traceback

# Types
from typing import List, cast
from mindgard.types import log_levels, model_types, valid_image_datasets, type_model_presets_list

# Models
from mindgard.preflight import preflight_image, preflight_llm
from mindgard.wrappers.utils import parse_args_into_model
from mindgard.wrappers.image import ImageModelWrapper
from mindgard.wrappers.llm import LLMModelWrapper

# Run functions
from mindgard.run_functions.list_tests import list_test_submit, list_test_polling, list_test_output
from mindgard.run_functions.sandbox_test import submit_sandbox_submit_factory, submit_sandbox_polling
from mindgard.run_functions.external_models import model_test_polling, model_test_output_factory, model_test_submit_factory
from mindgard.external_model_handlers.image_model import image_message_handler
from mindgard.external_model_handlers.llm_model import llm_message_handler
from mindgard.run_poll_display import cli_run

from mindgard.orchestrator import OrchestratorSetupRequest

# Constants and Utils
from mindgard.constants import VERSION
from mindgard.utils import is_version_outdated, print_to_stderr, parse_toml_and_args_into_final_args, convert_test_to_cli_response, CliResponse

# Logging
import logging
from rich.logging import RichHandler
from rich.console import Console

# Auth
from mindgard.auth import login, logout

# both validate and test need these same arguments, so have factored them out
def shared_arguments(parser: argparse.ArgumentParser):
    parser.add_argument('target', nargs='?', type=str, help="This is your own model identifier.")
    parser.add_argument('--config-file', type=str, help='Path to mindgard.toml config file', default=None, required=False)
    parser.add_argument('--json', action="store_true", help='Output the info in JSON format.', required=False, default=False)
    parser.add_argument('--headers', type=str, help='The headers to use. Comma separated list.', required=False)
    parser.add_argument('--header', type=str, help='The headers to use, repeat flag for each header.', action='append', required=False)
    parser.add_argument('--preset', type=str, help='The preset to use', choices=type_model_presets_list, required=False)
    parser.add_argument('--api-key', type=str, help='Specify the API key for the wrapper', required=False)
    parser.add_argument('--url', type=str, help='Specify the url for the wrapper', required=False)
    parser.add_argument('--model-name', type=str, help='Specify which model to run against (OpenAI and Anthropic)', required=False)
    parser.add_argument('--az-api-version', type=str, help='Specify the Azure OpenAI API version (Azure only)', required=False)
    parser.add_argument('--prompt', type=str, help='Specify the prompt to use', required=False)
    parser.add_argument('--system-prompt', type=str, help='Text file containing system prompt to use.', required=False)
    parser.add_argument('--selector', type=str, help='The selector to retrieve the text response from the LLM response JSON.', required=False)
    parser.add_argument('--request-template', type=str, help='The template to wrap the API request in.', required=False)
    parser.add_argument('--tokenizer', type=str, help='Choose a HuggingFace model to provide a tokeniser for prompt and chat completion templating.', required=False)
    parser.add_argument('--model-type', type=str, help='The modality of the model; image or llm', choices=model_types, required=False)
    parser.add_argument('--risk-threshold', type=int, help='Set a risk threshold above which the system will exit 1', required=False)

def parse_args(args: List[str]) -> argparse.Namespace:
    default_log_level = 'warn'

    parser = argparse.ArgumentParser(description='Securing AIs', prog='mindgard', usage='%(prog)s [command] [options]', epilog='Enjoy the program! :)', add_help=True)
    parser.add_argument('--version', action='version', version=f"%(prog)s {VERSION}", help='Show the current version number')
    parser.add_argument('--log-level', type=str, help='Specify the output verbosity', choices=log_levels, required=False, default=default_log_level)

    subparsers = parser.add_subparsers(dest='command', title='commands', description='Use these commands to interact with the Mindgard API')
    login_parser = subparsers.add_parser('login', help='Login to the Mindgard platform')
    login_parser.add_argument('--instance', nargs='?', type=str, help='Point to your deployed Mindgard instance. If not provided, cli will point towards Mindgard Sandbox')
    
    subparsers.add_parser('logout', help='Logout of the Mindgard platform in the CLI')

    sandbox_test_parser = subparsers.add_parser('sandbox', help='Test a mindgard example model')
    sandbox_test_parser.add_argument('target', nargs='?', type=str, choices=['cfp_faces', 'mistral'], default="cfp_faces")
    sandbox_test_parser.add_argument('--json', action="store_true", help='Return json output', required=False)
    sandbox_test_parser.add_argument('--risk-threshold', type=int, help='Set a risk threshold above which the system will exit 1', required=False, default=80)

    list_parser = subparsers.add_parser('list', help='List items')
    list_subparsers = list_parser.add_subparsers(dest='list_command')
    list_test_parser = list_subparsers.add_parser('tests', help='List tests')
    list_test_parser.add_argument('--json', action="store_true", help='Return json output', required=False)
    list_test_parser.add_argument('--id', type=str, help='Get the details of a specific test.', required=False)

    test_parser = subparsers.add_parser("test", help="Attacks command")
    shared_arguments(test_parser)
    test_parser.add_argument('--parallelism', type=int, help='The maximum number of parallel requests that can be made to the API.', required=False)
    test_parser.add_argument('--dataset', type=str, help='The dataset to use for image models', choices=valid_image_datasets, required=False)
    test_parser.add_argument('--mode', type=str, help='Specify the number of samples to use during attacks; contact Mindgard for access to \'thorough\' test', choices=['fast', 'thorough'], required=False)

    validate_parser = subparsers.add_parser("validate", help="Validates that we can communicate with your model")
    shared_arguments(validate_parser)

    return parser.parse_args(args)


def main() -> None:
    args = parse_args(sys.argv[1:])

    FORMAT = "%(message)s"
    logging.basicConfig(
        level=args.log_level.upper(), format=FORMAT, datefmt="[%X]", handlers=[RichHandler(console=Console(stderr=True))]
    )

    if not (sys.version_info.major == 3 and sys.version_info.minor >= 8):
        print_to_stderr("Python 3.8 or later is required to run the Mindgard CLI.")
        sys.exit(2)

    if new_version := is_version_outdated():
        print_to_stderr(f"New version available: {new_version}. Please upgrade as older versions of the CLI may not be actively maintained.")

    if args.command == 'login':
        login(instance=args.instance)
    elif args.command == 'logout':
        logout()
    elif args.command == 'list':
        if args.list_command == 'tests':
            cli_run(submit_func=list_test_submit, polling_func=list_test_polling, output_func=list_test_output, json_out=args.json, submitting_text="Fetching tests...")
            exit(CliResponse(0).code())
        else:
            print_to_stderr('Provide a resource to list. Eg `list tests`.')
    elif args.command == 'sandbox':
        submit_sandbox_submit = submit_sandbox_submit_factory(model_name=args.target)
        submit_sandbox_output = model_test_output_factory(risk_threshold=100)

        cli_response = cli_run(submit_func=submit_sandbox_submit, polling_func=submit_sandbox_polling, output_func=submit_sandbox_output, json_out=args.json)
        exit(convert_test_to_cli_response(test=cli_response, risk_threshold=100).code())

    elif args.command == "validate" or args.command == "test":
        console = Console()
        final_args = parse_toml_and_args_into_final_args(args.config_file, args)
        model_wrapper = parse_args_into_model(final_args["model_type"], final_args)

        if final_args["model_type"] == "llm":
            passed_preflight = preflight_llm(model_wrapper, console=console, json_out=final_args["json"])
        else:
            passed_preflight = preflight_image(model_wrapper, console=console, json_out=final_args["json"])

        if not final_args["json"]:
            console.print(f"{'[green bold]Model contactable!' if passed_preflight else '[red bold]Model not contactable!'}")


        if passed_preflight:
            if args.command == 'test':
                if final_args["model_type"] == "llm":
                    request = OrchestratorSetupRequest(
                        target=final_args["target"],
                        parallelism=int(final_args["parallelism"]),
                        system_prompt=final_args["system_prompt"],
                        dataset=None,
                        modelType=final_args["model_type"],
                        attackSource="user",
                        attackPack=("large" if final_args['mode'] == "thorough" else "sandbox"),
                    )
                    submit = model_test_submit_factory(
                        request=request,
                        model_wrapper=cast(LLMModelWrapper, model_wrapper),
                        message_handler=llm_message_handler
                    )
                elif final_args["model_type"] == "image":
                    dataset = final_args["dataset"] or valid_image_datasets[0]

                    request = OrchestratorSetupRequest(
                        target=final_args["target"],
                        parallelism=int(final_args["parallelism"]),
                        dataset=dataset,
                        modelType=final_args["model_type"],
                        attackSource="user",
                        labels=final_args["labels"]
                    )
                    submit = model_test_submit_factory(
                        request=request,
                        model_wrapper=cast(ImageModelWrapper, model_wrapper),
                        message_handler=image_message_handler
                    )

                output = model_test_output_factory(risk_threshold=int(final_args["risk_threshold"]))
                cli_response = cli_run(submit, model_test_polling, output_func=output, json_out=final_args["json"])
                exit(convert_test_to_cli_response(test=cli_response, risk_threshold=int(final_args["risk_threshold"])).code()) # type: ignore

        else:
            exit(CliResponse(1).code())
        
    else:
        print_to_stderr('Which command are you looking for? See: $ mindgard --help')


if __name__ == '__main__':
    try:
        main()
    except ValueError as e:
        print_to_stderr(str(e))
        exit(2)
    except Exception:
        traceback.print_exc()
        exit(2)


# OpenAI exceptions
from openai import (
    APIError,
    OpenAIError,
    ConflictError,
    NotFoundError,
    APIStatusError,
    RateLimitError,
    APITimeoutError,
    BadRequestError,
    APIConnectionError,
    AuthenticationError,
    InternalServerError as OPAIInternalServerError,
    PermissionDeniedError,
    UnprocessableEntityError,
    APIResponseValidationError,
)

# Typing
from typing import Dict
from requests import status_codes


class MGException(Exception):
    def __init__(self, message: str = "exception") -> None:
        super().__init__(self)
        self.message = message

    def __str__(self) -> str:
        return self.message


# Other Exceptions
class Uncontactable(MGException):
    pass


# HTTP-like Exceptions
class HTTPBaseError(MGException):
    def __init__(self, message: str, status_code: int) -> None:
        super().__init__(message)
        self.status_code: int = status_code
        self.status_message: str = "<unknown>"
        try:
            self.status_message = status_codes._codes[status_code][0]
        except:
            pass


class BadRequest(HTTPBaseError):
    pass


class Unauthorized(HTTPBaseError):
    pass


class Forbidden(HTTPBaseError):
    pass


class NotFound(HTTPBaseError):
    pass


class Timeout(HTTPBaseError):
    pass


class UnprocessableEntity(HTTPBaseError):
    pass


class FailedDependency(HTTPBaseError):
    pass


class RateLimitOrInsufficientCredits(HTTPBaseError):
    pass


class InternalServerError(HTTPBaseError):
    pass


class ServiceUnavailable(HTTPBaseError):
    pass


class EmptyResponse(MGException):
    pass


class NotImplemented(MGException):
    pass


_status_code_exception_map: Dict[int, HTTPBaseError] = {
    400: BadRequest("LLM provider received message that couldn't be handled.", 400),
    401: Unauthorized(
        "User is not authorized to access this LLM provider resource.", 401
    ),
    403: Forbidden(
        "User is known but does not have permission to access this resource.", 403
    ),
    404: NotFound("This resource was not found.", 404),
    408: Timeout("Timed out while trying to access resource", 408),
    422: UnprocessableEntity("Entity sent to LLM could not be processed.", 422),
    424: FailedDependency("Failed Dependency TODO", 424),
    429: RateLimitOrInsufficientCredits("Rate Limit or Insufficient Credits TODO", 429),
    500: InternalServerError("Internal Server Error TODO", 500),
    503: ServiceUnavailable("Service Unavailable TODO", 503),
}


def status_code_to_exception(status_code: int) -> HTTPBaseError:
    return _status_code_exception_map.get(
        status_code, HTTPBaseError("Error specifics unknown", -1)
    )


def openai_exception_to_exception(exception: OpenAIError) -> MGException:
    # if its an API error with a status code do this
    if isinstance(exception, APIStatusError):
        if exception.status_code == 422:
            err_message = "<none>"
            try:
                err_message = exception.response.json().get("error", err_message)
            except Exception:
                pass
            return UnprocessableEntity(
                f"Received 422 from provider: {err_message}", 422
            )
        return status_code_to_exception(exception.status_code)
    # otherwise do this, this needs to be refined more
    else:
        return EmptyResponse("response was empty or something")
